@@ -64,8 +64,7 @@ gboolean file_changed = FALSE;
 static guint play_id;
 static guint record_id;
 
-///////////////////////////////////////////////////////////////////
-// ------------------- Callbacks ------------------------------- //
+/* ------------------- Callbacks ------------------------------- */
 void
 on_record_activate_cb (GtkWidget* widget, gpointer data)
 {
@@ -76,6 +75,9 @@ on_record_activate_cb (GtkWidget* widget, gpointer data)
 	grecord_set_sensitive_progress ();
 	file_changed = TRUE;
 
+	/* Reset record-time and stuff */
+	UpdateStatusbarRecord (TRUE);
+
 	RecEng.pid = fork ();
 	if (RecEng.pid == 0) {
 		
@@ -92,7 +94,7 @@ on_record_activate_cb (GtkWidget* widget, gpointer data)
 
 	record_id = gtk_timeout_add (1000,
 				     (GtkFunction) UpdateStatusbarRecord,
-				     NULL);
+				     FALSE);
 }
 
 void
@@ -506,8 +508,7 @@ on_undoall_activate_cb (GtkWidget* widget, gpointer data)
 	g_free (temp_string);
 }
 
-///////////////////////////////////////////////////////////////
-// --------------------- Help functions -------------------- //
+/* --------------------- Help functions -------------------- */
 void
 record_sound (void)
 {
@@ -713,7 +714,7 @@ save_filename (GtkFileSelection* selector, gpointer file_selector)
 		gchar* tempstring;
 		gchar* string = g_strdup ((char *) strrchr (active_file, '/'));
 		string[1] = ' ';
-		temp_string = g_strconcat (_(maintopic), string, NULL);
+		tempstring = g_strconcat (_(maintopic), string, NULL);
 		gtk_window_set_title (GTK_WINDOW (grecord_widgets.grecord_window), tempstring);
 		g_free (tempstring);
 		g_free (string);
@@ -819,12 +820,19 @@ UpdateStatusbarPlay (gboolean begin)
 }
 
 guint
-UpdateStatusbarRecord (gpointer data)
+UpdateStatusbarRecord (gboolean begin)
 {
 	static gint counter = 0;
 	static gint timeout = 0;
 	static gint countersec = 0;
 
+	if (begin) {
+		counter = 0.00;
+		timeout = 0;
+		countersec = 0;
+		return TRUE;
+	}
+
 	/* Timeout */
 	if (counter >= 1000) {
 		if (stop_on_timeout) {
