@@ -39,6 +39,7 @@
 #include <errno.h>
 #include <math.h>
 #include <sys/ioctl.h>
+#include <ctype.h>
 
 /* For _() */
 #include <config.h>
@@ -98,6 +99,8 @@ int tcd_init_disc( cd_struct *cd, WarnFunc msg_cb )
     cd->nslots = 0;
 #endif
     cd->old_cddb_id = 0;
+
+    tcd_close_disc( cd );
     
     debug("cdrom.c: tcd_init_disc exiting normally\n" );
     return(tcd_post_init(cd));
@@ -129,18 +132,21 @@ int tcd_close_disc( cd_struct *cd )
 {
     debug("cdrom.c: tcd_close_disc(%p) top\n", cd );
     close(cd->cd_dev);
+    cd->cd_dev = -1;
     debug("cdrom.c: tcd_close_disc exiting normally\n" );
     return 0;
 }
 
-int tcd_readtoc( cd_struct *cd )
+void tcd_readtoc( cd_struct *cd )
 {
-    int tmp,i;
+    int i;
     int delsecs;
 
     if(cd->time_lock)
 	    return;
 
+    tcd_opencddev( cd, NULL );
+
     debug("cdrom.c: tcd_readtoc(%p) top\n", cd );
     cd->err = FALSE;
     cd->isplayable=FALSE;
@@ -153,7 +159,8 @@ int tcd_readtoc( cd_struct *cd )
 	debug("cdrom.c: tcd_readtoc exiting prematurly. CDROMREADTOCHDR ioctl error.\n" );
 	cd->cur_t = 0;
 	cd->cddb_id = 0;
-	return(-1);		
+	tcd_close_disc ( cd );
+	return;
     }
 
     /* grab first & last tracks */
@@ -173,7 +180,8 @@ int tcd_readtoc( cd_struct *cd )
 	debug("cdrom.c: tcd_readtoc exiting prematurly. CDROMREADTOCENTRY ioctl error.\n" );
 	cd->cur_t = 0;
 	cd->cddb_id = 0;
-	return(-1);
+	tcd_close_disc ( cd );
+	return;
     }                                         
 
     /* read the rest of the tocs */
@@ -189,7 +197,8 @@ int tcd_readtoc( cd_struct *cd )
 	    debug("cdrom.c: tcd_readtoc exiting prematurly. CDROMREADTOCENTRY ioctl error.\n" );
 	    cd->cur_t = 0;
 	    cd->cddb_id = 0;
-	    return(-1);
+	    tcd_close_disc ( cd );
+	    return;
 	}
 
 	cd->trk[C(i)].type = cd->trk[C(i)].toc.cdte_ctrl;
@@ -224,13 +233,14 @@ int tcd_readtoc( cd_struct *cd )
     cd->cddb_id = cddb_discid(cd);
 
     cd->isplayable=TRUE;
+    tcd_close_disc ( cd );
     debug("cdrom.c: tcd_readtoc exiting normally\n" );
-    return tmp;
+    return;
 }
 
 void tcd_recalculate(cd_struct *cd)
 {
-    int result;
+    int result=0;
     if(cd->time_lock)
 	    return;
 
@@ -256,13 +266,15 @@ void tcd_recalculate(cd_struct *cd)
     cd->cd_min = cd->cur_pos_abs / 60;
 
 #ifdef TCD_CHANGER_ENABLED
+    tcd_opencddev( cd, NULL );
     cd->cur_disc = ioctl( cd->cd_dev, CDROM_SELECT_DISC, CDSL_CURRENT );
+    tcd_close_disc ( cd );
 #endif
 }
 
 void tcd_recalculate_fake(cd_struct *cd, gint abs_pos, gint track)
 {
-    int result;
+    int result=0;
 
     /* calculate various timing values */
     cd->cur_t = track;
@@ -286,19 +298,20 @@ void tcd_recalculate_fake(cd_struct *cd, gint abs_pos, gint track)
     cd->cd_min = cd->cur_pos_abs / 60;
 }
 
-/* finds the track the current second in teh album is in */
+/* finds the track the current second in the album is in */
 int tcd_find_track(cd_struct *cd, gint abs_pos)
 {
 	int t;
 	
-	for(t = cd->first_t; t < cd->last_t;)
+	for(t = cd->first_t; t <= cd->last_t; t++)
 	{
-		if(abs_pos >= cd->trk[t].start/75)
-			t++;
-		else
+		if(cd->trk[t].toc.cdte_ctrl == CDROM_DATA_TRACK)
 			return t-1;
+		if(abs_pos >= cd->trk[t].start/75)
+			continue;
+		return t-1;
 	}
-	return t;
+	return cd->last_t;
 }
 
 void tcd_gettime( cd_struct *cd )
@@ -308,7 +321,11 @@ void tcd_gettime( cd_struct *cd )
 	
 	if(cd->isplayable)
 	{
-		if(ioctl( cd->cd_dev, CDROMSUBCHNL, &cd->sc))
+	        int tmp;
+		tcd_opencddev (cd, NULL);
+		tmp = ioctl( cd->cd_dev, CDROMSUBCHNL, &cd->sc);
+		tcd_close_disc (cd);
+		if(tmp)
 		{
 			strcpy( cd->errmsg, "Can't read disc." );
 			cd->err = TRUE;
@@ -327,11 +344,15 @@ void tcd_gettime( cd_struct *cd )
 int tcd_set_volume(cd_struct *cd, int volume)
 {
     struct cdrom_volctrl vol;
+    int tmp;
 
     vol.channel0 = volume;
     vol.channel1 = vol.channel2 = vol.channel3 = vol.channel0;
-	
-    if(ioctl(cd->cd_dev, CDROMVOLCTRL, &vol) < 0)
+
+    tcd_opencddev( cd, NULL );
+    tmp = ioctl(cd->cd_dev, CDROMVOLCTRL, &vol);
+    tcd_close_disc ( cd );
+    if(tmp < 0)
 	return FALSE;
 
     return TRUE;
@@ -341,8 +362,12 @@ int tcd_get_volume(cd_struct *cd)
 {
 #ifdef CDROMVOLREAD
     struct cdrom_volctrl vol;
+    int tmp;
 
-    if(ioctl(cd->cd_dev, CDROMVOLREAD, &vol) < 0)
+    tcd_opencddev( cd, NULL );
+    tmp = ioctl(cd->cd_dev, CDROMVOLREAD, &vol);
+    tcd_close_disc ( cd );
+    if(tmp < 0)
 	return -1;
 
     return vol.channel0;
@@ -351,15 +376,17 @@ int tcd_get_volume(cd_struct *cd)
 #endif
 }	
 	                                  
-int tcd_playtracks(cd_struct *cd, int start_t, int end_t, int only_use_trkind)
+void tcd_playtracks(cd_struct *cd, int start_t, int end_t, int only_use_trkind)
 {
     struct cdrom_msf msf;
     struct cdrom_ti trkind;
-    int tmp;
     debug("cdrom.c: tcd_playtracks( %p, %d, %d )\n", cd, start_t, end_t );
     cd->err = FALSE;
 	
     /* make sure we can play it */
+    if(!cd->isplayable)
+	return;
+
     tcd_gettime(cd);
     if(cd->err) 
     {
@@ -369,10 +396,11 @@ int tcd_playtracks(cd_struct *cd, int start_t, int end_t, int only_use_trkind)
 	if(cd->err) 
 	{
 	    debug("cdrom.c: tcd_playtracks - error while fetching disc.\n");
-	    return -1;
+	    return;
 	}
     }
 
+    tcd_opencddev( cd, NULL );
 #if defined(CDROMCLOSETRAY)
     if( ioctl( cd->cd_dev, CDROM_DRIVE_STATUS ) == CDS_TRAY_OPEN )
 	    ioctl(cd->cd_dev, CDROMCLOSETRAY);
@@ -433,13 +461,15 @@ int tcd_playtracks(cd_struct *cd, int start_t, int end_t, int only_use_trkind)
 	    strcpy( cd->errmsg, "Error playing disc" );
 	    cd->err = TRUE;
 	    debug("cdrom.c: tcd_playtracks error. CDROMPLAYTRKIND ioctl error.\n");
-	    return -1;
+	    tcd_close_disc ( cd );
+	    return;
 	}
     }
+    tcd_close_disc ( cd );
 
     cd->isplayable = TRUE;
     debug("cdrom.c: tcd_playtracks exiting normally\n" );
-    return tmp;
+    return;
 }       
 
 static int msf_2_frame( cd_min_sec_frame *msf )
@@ -461,6 +491,7 @@ int tcd_play_seconds( cd_struct *cd, long int offset )
     struct cdrom_msf msf;
     cd_min_sec_frame msf0;
     int cur_frame, start_frame, end_frame;
+    int tmp;
 
     debug("cdrom.c: tcd_play_seconds( %p, %ld )\n", cd, offset );
 
@@ -506,18 +537,20 @@ int tcd_play_seconds( cd_struct *cd, long int offset )
     }
 #endif
 	
-    if(ioctl(cd->cd_dev, CDROMPLAYMSF, &msf))
-    {
+    tcd_opencddev( cd, NULL );
+    tmp = ioctl(cd->cd_dev, CDROMPLAYMSF, &msf);
+    if(tmp)
+      {
 	strcpy( cd->errmsg, "Error playing disc." );
 	cd->err = TRUE;
 
 	debug("cdrom.c: tcd_play_seconds error. CDROMPLAYMSF ioctl error.\n" );
-	return(-1);
     }
-    cd->isplayable=TRUE;                                                 
+    cd->isplayable=TRUE;
+    tcd_close_disc( cd );
 
     debug("cdrom.c: tcd_play_seconds exiting normally\n" );
-    return( 0 );
+    return tmp ? -1 : 0;
 }       
 
 int tcd_ejectcd( cd_struct *cd )
@@ -528,6 +561,7 @@ int tcd_ejectcd( cd_struct *cd )
     if(cd->isplayable) tcd_stopcd(cd);
     cd->err = FALSE;
 
+    tcd_opencddev( cd, NULL );
     if(!ioctl(cd->cd_dev, CDROMEJECT))
     {
 	cd->isplayable = FALSE;
@@ -552,7 +586,8 @@ int tcd_ejectcd( cd_struct *cd )
 	}
 	cd->isplayable = TRUE;
     }
-    cd->cur_t = 0;	
+    cd->cur_t = 0;
+    tcd_close_disc( cd );
 	
     debug("cdrom.c: tcd_eject exiting normally\n" );
     return 0;
@@ -569,7 +604,10 @@ int tcd_stopcd(cd_struct *cd)
 	tcd_pausecd(cd);
 
     cd->err = FALSE;
-    if(ioctl(cd->cd_dev, CDROMSTOP))
+    tcd_opencddev( cd, NULL );
+    tmp = ioctl(cd->cd_dev, CDROMSTOP);
+    tcd_close_disc ( cd );
+    if(tmp)
     {
 	strcpy( cd->errmsg, "Can't stop disc." );
 	cd->err = TRUE;
@@ -587,26 +625,23 @@ int tcd_pausecd( cd_struct *cd )
     int tmp;
     cd->err = FALSE;
 	
+    tcd_opencddev( cd, NULL );
     if(cd->sc.cdsc_audiostatus==CDROM_AUDIO_PAUSED)
-    {       
-	if((tmp=ioctl(cd->cd_dev, CDROMRESUME)))
-	{
-	    strcpy(cd->errmsg, strerror(errno));
-	    cd->err = TRUE;
-	    return(-1);
-	}
-	return tmp;
+    {
+        tmp = ioctl(cd->cd_dev, CDROMRESUME);
     }	        
     else
     {
-	if((tmp=ioctl(cd->cd_dev, CDROMPAUSE)))
-	{
-	    strcpy( cd->errmsg, strerror( errno ) );
-	    cd->err = TRUE;
-	    return(-1);
-	}
-	return tmp;
+        tmp=ioctl(cd->cd_dev, CDROMPAUSE);
     }
+    if(tmp < 0)
+    {
+        strcpy(cd->errmsg, strerror(errno));
+	cd->err = TRUE;
+    }
+    tcd_close_disc ( cd );
+
+    return tmp ? -1 : 0;
 }
 
 int tcd_change_disc( cd_struct *cd, int disc )
@@ -615,7 +650,9 @@ int tcd_change_disc( cd_struct *cd, int disc )
     int tmp;
     cd->err = FALSE;
 
+    tcd_opencddev( cd, NULL );
     tmp = ioctl( cd->cd_dev, CDROM_SELECT_DISC, disc );
+    tcd_close_disc ( cd );
     if(tmp && errno)
 	fprintf( stdout, "ioctl: %s\n", strerror(errno) );	
 
@@ -649,22 +686,46 @@ void tcd_opencddev( cd_struct *cd, WarnFunc msg_cb )
     }
 }
 
+/*
+ * Parse the title information for a CD.  This information is in the
+ * format "Artist / Album" where the spaces around the divider may or
+ * may not be included.
+ */
 void parse_dtitle(cd_struct *cd)
 {
     char tmp[DISC_INFO_LEN], *tmp2;
+    int len;
 
-    /* Parse out the individual title elements. Help from Alec M. */
+    /* Parse out the individual title elements. */
     strncpy(tmp, cd->dtitle, DISC_INFO_LEN);
-    if (tmp[0])
-	strncpy(cd->artist, strtok(tmp, "/"), DISC_INFO_LEN);
-    else
-	cd->artist[0]=0;
+    if (tmp[0] == '\0') {
+	cd->artist[0] = '\0';
+	cd->album[0] = '\0';
+	return;
+    }
 
-    tmp2 = strtok(NULL, "\0");
-    if(tmp2)
-	strncpy(cd->album, tmp2+1, DISC_INFO_LEN);
+    tmp2 = index(tmp, '/');
+    if (tmp2)
+    {
+	len = tmp2 - tmp;
+	if (len > 0)
+	{				/* Back up to first non-space */
+	    while ((len > 0) && isspace(tmp[len-1]))
+		   len--;
+	}
+	strncpy(cd->artist, tmp, len);
+	cd->artist[len] = '\0';		/* Tie off the string */
+
+	tmp2++;				/* Skip the '/' */
+	while (isspace(*tmp2))		/* Skip spaces */
+	    tmp2++;
+	strncpy(cd->album, tmp2, DISC_INFO_LEN);
+    }
     else
-	strncpy(cd->album, "", DISC_INFO_LEN);
+    {
+	strncpy(cd->artist, tmp, DISC_INFO_LEN);
+	cd->album[0] = '\0';
+    }
 }
 
 int tcd_readdiskinfo( cd_struct *cd )
