@@ -26,6 +26,7 @@
 #include <gnome.h>
 #include <string.h>
 #include <ctype.h>
+#include <libgnomeui/gnome-window-icon.h>
 
 #include "gtcd_public.h"
 #include "keybindings.h"
@@ -58,6 +59,120 @@ void save_prefs(tcd_prefs *prop);
 void changed_cb(GtkWidget *widget, void *data);
 void preferences(GtkWidget *widget, void *data);
 
+/*
+ * Convert a shortcut data structure into a user readable format.
+ *
+ * @param buf A pointer to the buffer where the data should be
+ * written.
+ *
+ * @param len The length of this buffer
+ *
+ * @param key The key to be printed.
+ *
+ * @return gchar* Returns the buffer passed in as the first argument.
+ * This allows the routine to be used within another call to printf().
+ */
+static gchar *print_key (gchar *buf, int len, Shortcut *key)
+{
+	g_snprintf(buf, len, "%s%s%s%s%s%s%c",
+		   (key->mods & GDK_CONTROL_MASK) ? "C-"  : "",
+		   (key->mods & GDK_MOD1_MASK)    ? "M1-" : "",
+		   (key->mods & GDK_MOD2_MASK)    ? "M2-" : "",
+		   (key->mods & GDK_MOD3_MASK)    ? "M3-" : "",
+		   (key->mods & GDK_MOD4_MASK)    ? "M4-" : "",
+		   (key->mods & GDK_MOD5_MASK)    ? "M5-" : "",
+		   toupper(key->key));
+	return(buf);
+}
+
+/*
+ * Save a shortcut to the configuration file.  This converts the
+ * shortcut from its internal format to a user readable format and
+ * then puts it into the program's configuration file.
+ *
+ * @param gconf_key A pointer to the string containing the gconf key.
+ *
+ * @param key A pointer to the internal storage for the key.
+ */
+static void save_key(gchar *gconf_key, Shortcut *key)
+{
+	gchar value[32];
+
+	print_key(value, 32, key);
+	gnome_config_set_string(gconf_key, value);
+
+}
+
+/*
+ * Check the ASCII key representation for a leading string.  If that
+ * string is found then add the appropriate flag to the key modifiers.
+ * This routine is called repeatedly to check for each possible
+ * modifier.
+ *
+ * @param buf A pointer to the current location in the ASCII key
+ * representation.
+ *
+ * @param key A pointer to the internal storage for the key.
+ *
+ * @param pattern The pattern to test.
+ *
+ * @param mask The flag to set if the pattern is matched.
+ *
+ * @return gchar* A pointer to the next location to check in the ASCII
+ * key string.
+ */
+static gchar *check_for_mod (gchar *buf, Shortcut *key, gchar *pattern, guint mask)
+{
+	gint len;
+
+	len = strlen(pattern);
+	if (strncmp(buf, pattern, len) == 0)
+	{
+		key->mods |= mask;
+		buf += len;
+	}
+	return(buf);
+}
+
+/*
+ * Read a shortcut from the configuration file.  This converts the
+ * shortcut from a user readable format to its internal format.
+ *
+ * @param key A pointer to the internal storage for the key.
+ *
+ * @param gconf_key A pointer to the string containing the gconf key.
+ */
+static void load_key(Shortcut *key, gchar *gconf_key)
+{
+	gchar *value, *c;
+
+	value = gnome_config_get_string(gconf_key);
+
+	key->mods = 0;
+	c = check_for_mod(value, key, "C-",  GDK_CONTROL_MASK);
+	c = check_for_mod(c,     key, "M1-", GDK_MOD1_MASK);
+	c = check_for_mod(c,     key, "M2-", GDK_MOD2_MASK);
+	c = check_for_mod(c,     key, "M3-", GDK_MOD3_MASK);
+	c = check_for_mod(c,     key, "M4-", GDK_MOD4_MASK);
+	c = check_for_mod(c,     key, "M5-", GDK_MOD5_MASK);
+	key->key = c[0];
+	free(value);
+}
+
+/*
+ * Read an old shortcut from the configuration file.  This converts
+ * the shortcut from a user readable format to its internal format.
+ *
+ * @param key A pointer to the internal storage for the key.
+ *
+ * @param gconf_key A pointer to the string containing the gconf key.
+ */
+static void load_old_key(Shortcut *key, gchar *gconf_key)
+{
+	key->mods = 0;
+	key->key = gnome_config_get_int(gconf_key);
+}
+
 void load_prefs(tcd_prefs *prop)
 {
 	char *s;
@@ -74,7 +189,7 @@ void load_prefs(tcd_prefs *prop)
 	prop->tooltip=gnome_config_get_bool    ("/gtcd/ui/tooltip=true");
 	prop->mixer_cmd=gnome_config_get_string    ("/gtcd/ui/mixer=gmix");
 
-	s = g_strconcat("/gtcd/ui/trackfont=", _("-adobe-helvetica-medium-r-normal-*-*-120-*-*-p-*-*-*"), NULL);
+	s = g_strconcat("/gtcd/ui/trackfont=", DEFAULT_FONT, NULL);
 	prop->trackfont=gnome_config_get_string(s);
 	g_free(s);
 
@@ -86,14 +201,33 @@ void load_prefs(tcd_prefs *prop)
 	prop->start_action=gnome_config_get_int        ("/gtcd/general/start_action=0");
 	prop->close_tray_on_start=gnome_config_get_bool("/gtcd/general/close_tray_on_start=false");
 
-	prop->quit.key =    gnome_config_get_int("/gtcd/keybindings/quit=81");
-	prop->play.key =    gnome_config_get_int("/gtcd/keybindings/play=80");
-	prop->stop.key =    gnome_config_get_int("/gtcd/keybindings/stop=83");
-	prop->tracked.key = gnome_config_get_int("/gtcd/keybindings/tracked=84");
-	prop->mixer.key =   gnome_config_get_int("/gtcd/keybindings/mixer=109");
-	prop->eject.key =   gnome_config_get_int("/gtcd/keybindings/eject=69");
-	prop->back.key =    gnome_config_get_int("/gtcd/keybindings/back=45");
-	prop->forward.key = gnome_config_get_int("/gtcd/keybindings/forward=43");
+	load_key(&prop->quit,         "/gtcd/new_keybindings/quit=Q");
+	load_key(&prop->play,         "/gtcd/new_keybindings/play=P");
+	load_key(&prop->stop,         "/gtcd/new_keybindings/stop=S");
+	load_key(&prop->tracked,      "/gtcd/new_keybindings/tracked=T");
+	load_key(&prop->mixer,        "/gtcd/new_keybindings/mixer=M");
+	load_key(&prop->eject,        "/gtcd/new_keybindings/eject=E");
+	load_key(&prop->back,         "/gtcd/new_keybindings/back=-");
+	load_key(&prop->forward,      "/gtcd/new_keybindings/forward=+");
+	load_key(&prop->preferences,  "/gtcd/new_keybindings/preferences=C-P");
+	load_key(&prop->fast_forward, "/gtcd/new_keybindings/fast_forward=]");
+	load_key(&prop->rewind,       "/gtcd/new_keybindings/rewind=[");
+	load_key(&prop->vol_up,       "/gtcd/new_keybindings/volume_up=>");
+	load_key(&prop->vol_down,     "/gtcd/new_keybindings/volume_down=<");
+
+	if (gnome_config_has_section("/gtcd/keybindings"))
+	{
+		/* One time event: override with any old keybindings */
+	    load_old_key(&prop->quit,    "/gtcd/keybindings/quit=81");
+	    load_old_key(&prop->play,    "/gtcd/keybindings/play=80");
+	    load_old_key(&prop->stop,    "/gtcd/keybindings/stop=83");
+	    load_old_key(&prop->tracked, "/gtcd/keybindings/tracked=84");
+	    load_old_key(&prop->mixer,   "/gtcd/keybindings/mixer=109");
+	    load_old_key(&prop->eject,   "/gtcd/keybindings/eject=69");
+	    load_old_key(&prop->back,    "/gtcd/keybindings/back=45");
+	    load_old_key(&prop->forward, "/gtcd/keybindings/forward=43");
+	    gnome_config_clean_section("/gtcd/keybindings");
+	}
 
 	prop->only_use_trkind = gnome_config_get_bool("/gtcd/general/only_use_trkind=0");
 	prop->squared_volume = gnome_config_get_bool("/gtcd/general/squared_volume=0");
@@ -126,14 +260,19 @@ void save_prefs(tcd_prefs *prop)
 	gnome_config_set_int ("/gtcd/general/start_action", prop->start_action);
 	gnome_config_set_bool("/gtcd/general/close_tray_on_start", prop->close_tray_on_start);
 
-	gnome_config_set_int("/gtcd/keybindings/quit", prop->quit.key);
-	gnome_config_set_int("/gtcd/keybindings/play", prop->play.key);
-	gnome_config_set_int("/gtcd/keybindings/stop", prop->stop.key);
-	gnome_config_set_int("/gtcd/keybindings/tracked", prop->tracked.key);
-	gnome_config_set_int("/gtcd/keybindings/mixer", prop->mixer.key);
-	gnome_config_set_int("/gtcd/keybindings/eject", prop->eject.key);
-	gnome_config_set_int("/gtcd/keybindings/back", prop->back.key);
-	gnome_config_set_int("/gtcd/keybindings/forward", prop->forward.key);
+	save_key("/gtcd/new_keybindings/quit", &prop->quit);
+	save_key("/gtcd/new_keybindings/play", &prop->play);
+	save_key("/gtcd/new_keybindings/stop", &prop->stop);
+	save_key("/gtcd/new_keybindings/tracked", &prop->tracked);
+	save_key("/gtcd/new_keybindings/mixer", &prop->mixer);
+	save_key("/gtcd/new_keybindings/eject", &prop->eject);
+	save_key("/gtcd/new_keybindings/back", &prop->back);
+	save_key("/gtcd/new_keybindings/forward", &prop->forward);
+	save_key("/gtcd/new_keybindings/preferences", &prop->preferences);
+	save_key("/gtcd/new_keybindings/fast_forward", &prop->fast_forward);
+	save_key("/gtcd/new_keybindings/rewind", &prop->rewind);
+	save_key("/gtcd/new_keybindings/volume_up", &prop->vol_up);
+	save_key("/gtcd/new_keybindings/volumn_down", &prop->vol_down);
 
 	gnome_config_set_string("/cddbslave/server/address", prop->cddb_server);
 	gnome_config_set_int("/cddbslave/server/port=8880", prop->cddb_port);
@@ -287,7 +426,9 @@ static GtkWidget *create_exit_frame()
 
 static void dev_entry_changed_cb(GtkWidget *widget, gpointer data)
 {
-	prefs->cddev = g_strdup(gtk_entry_get_text(GTK_ENTRY(widget)));
+	gchar **handle = data;
+
+	*handle = g_strdup(gtk_entry_get_text(GTK_ENTRY(widget)));
 	changed_cb(NULL, NULL);
 }
 
@@ -342,7 +483,16 @@ GtkWidget *create_general_frame()
 	dev_entry = gtk_entry_new();
 	gtk_entry_set_text(GTK_ENTRY(dev_entry), prefs->cddev);
 	gtk_signal_connect(GTK_OBJECT(dev_entry), "changed",
-			   GTK_SIGNAL_FUNC(dev_entry_changed_cb), NULL);
+			   GTK_SIGNAL_FUNC(dev_entry_changed_cb), &prefs->cddev);
+	gtk_box_pack_start_defaults(GTK_BOX(left_box), label);
+	gtk_box_pack_start_defaults(GTK_BOX(right_box), dev_entry);
+    
+	/* device entry */
+	label = gtk_label_new(_("Mixer Command"));
+	dev_entry = gtk_entry_new();
+	gtk_entry_set_text(GTK_ENTRY(dev_entry), prefs->mixer_cmd);
+	gtk_signal_connect(GTK_OBJECT(dev_entry), "changed",
+			   GTK_SIGNAL_FUNC(dev_entry_changed_cb), &prefs->mixer_cmd);
 	gtk_box_pack_start_defaults(GTK_BOX(left_box), label);
 	gtk_box_pack_start_defaults(GTK_BOX(right_box), dev_entry);
     
@@ -437,7 +587,7 @@ static void fill_list(KeyBinding *kb, GtkWidget *clist)
 	tmp[0] = g_malloc(64);	/* key */
 	tmp[1] = g_malloc(256);	/* desc */
 
-	g_snprintf(tmp[0], 63, "%c", toupper(kb->key->key));
+	print_key(tmp[0], 63, kb->key);
 	g_snprintf(tmp[1], 255, "%s", kb->desc);
     
 	gtk_clist_append(GTK_CLIST(clist), tmp);
@@ -452,11 +602,14 @@ static int entry_changed(GtkWidget *widget, GdkEvent *ev, KeyBinding *kb)
 	char tmp[64];
 	if(ev->type != GDK_KEY_PRESS)
 		return 0;
-	g_snprintf(tmp, 63, "%c", toupper(kb->key->key));
+	if (((GdkEventKey*)ev)->length == 0)
+		return 0;
+	print_key(tmp, 63, kb->key);
 	gtk_object_remove_data(GTK_OBJECT(kb->data), tmp);
-	kb->key->key = *((GdkEventKey*)ev)->string;
+	kb->key->key  = ((GdkEventKey*)ev)->keyval;
+	kb->key->mods = ((GdkEventKey*)ev)->state & GTCD_MOD_MASK;
 
-	g_snprintf(tmp, 63, "%c", toupper(kb->key->key));
+	print_key(tmp, 63, kb->key);
 	gtk_object_set_data(GTK_OBJECT(kb->data), tmp, kb);
 	gtk_clist_set_text(GTK_CLIST(kb->data), kb->data2, 0, tmp);
 	changed_cb(NULL, NULL);
@@ -470,8 +623,8 @@ static void select_row_cb(GtkCList *clist,
 			  GdkEventButton *event,
 			  gpointer data)
 {
-	static int entry_cb=0, alt_cb=0, shift_cb=0, ctrl_cb=0;
-	GtkWidget *entry, *alt_check, *shift_check, *ctrl_check;
+	static int entry_cb=0;
+	GtkWidget *entry;
 	KeyBinding *kb;
 	gchar *text;
 
@@ -480,29 +633,18 @@ static void select_row_cb(GtkCList *clist,
 	entry = gtk_object_get_data(GTK_OBJECT(clist), "entry");
 	if(!kb) /* FIXME this is broken...? */
 		return;
-	ctrl_check = gtk_object_get_data(GTK_OBJECT(clist), "ctrl_check");
-	alt_check = gtk_object_get_data(GTK_OBJECT(clist), "alt_check");
-	shift_check = gtk_object_get_data(GTK_OBJECT(clist), "shift_check");
 
 	if(entry_cb > 0) gtk_signal_disconnect(GTK_OBJECT(entry), entry_cb);
-	if(ctrl_cb > 0) gtk_signal_disconnect(GTK_OBJECT(ctrl_check), ctrl_cb);
-	if(alt_cb > 0) gtk_signal_disconnect(GTK_OBJECT(alt_check), alt_cb);
-	if(shift_cb > 0) gtk_signal_disconnect(GTK_OBJECT(shift_check), shift_cb);
+	if (GPOINTER_TO_INT(data))
+	{
+		entry_cb = 0;
+		return;
+	}
 
-	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(ctrl_check), kb->key->ctrl);
-	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(alt_check), kb->key->alt);
-	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(shift_check), kb->key->shift);
-    
 	kb->data = (GtkWidget*)clist;
 	kb->data2 = row;
 	entry_cb = gtk_signal_connect(GTK_OBJECT(entry), "event",
 				      GTK_SIGNAL_FUNC(entry_changed), kb);
-	ctrl_cb = gtk_signal_connect(GTK_OBJECT(ctrl_check), "clicked",
-				     GTK_SIGNAL_FUNC(check_changed_cb), &kb->key->ctrl);
-	alt_cb = gtk_signal_connect(GTK_OBJECT(alt_check), "clicked",
-				    GTK_SIGNAL_FUNC(check_changed_cb), &kb->key->alt);
-	shift_cb = gtk_signal_connect(GTK_OBJECT(shift_check), "clicked",
-				      GTK_SIGNAL_FUNC(check_changed_cb), &kb->key->shift);
 }
 
 static GtkWidget *key_page(void)
@@ -534,7 +676,10 @@ static GtkWidget *key_page(void)
 
 	gtk_container_add(GTK_CONTAINER(scrolled), clist);
 	gtk_signal_connect(GTK_OBJECT(clist), "select_row",
-			   GTK_SIGNAL_FUNC(select_row_cb), NULL);
+			   GTK_SIGNAL_FUNC(select_row_cb), FALSE);
+	gtk_signal_connect(GTK_OBJECT(clist), "destroy",
+			   GTK_SIGNAL_FUNC(select_row_cb),
+			   GINT_TO_POINTER(TRUE));
 	/* Box */
 	box = gtk_vbox_new(FALSE, GNOME_PAD_SMALL);
 	gtk_box_pack_start(GTK_BOX(box), scrolled, TRUE, TRUE, 0);
@@ -551,18 +696,10 @@ static GtkWidget *key_page(void)
 
 	gtk_box_pack_start(GTK_BOX(mod_box), label, FALSE, FALSE, GNOME_PAD);
 	gtk_box_pack_start(GTK_BOX(mod_box), entry, TRUE, TRUE, GNOME_PAD);
-#if 0				/* Not yet implemented */
-	gtk_box_pack_start_defaults(GTK_BOX(mod_box), ctrl_check);
-	gtk_box_pack_start_defaults(GTK_BOX(mod_box), alt_check);
-	gtk_box_pack_start_defaults(GTK_BOX(mod_box), shift_check);
-#endif
 
 	gtk_box_pack_start(GTK_BOX(box), mod_box, FALSE, FALSE, 0);
 
 	gtk_object_set_data(GTK_OBJECT(clist), "entry", entry);
-	gtk_object_set_data(GTK_OBJECT(clist), "ctrl_check", ctrl_check);
-	gtk_object_set_data(GTK_OBJECT(clist), "alt_check", alt_check);
-	gtk_object_set_data(GTK_OBJECT(clist), "shift_check", shift_check);
 
 	/* Frame */
 	frame = gtk_frame_new(_("Keybindings"));
